#![no_std]
#![no_main]

mod crypto;

use arduino_hal::hal::usart::BaudrateArduinoExt;
use arduino_hal::prelude::*;
use arduino_hal::{delay_ms, Usart};
use crypto::hmac_sha1;

use panic_halt as _;

const INTERVAL: u64 = 30;
const SECRET_KEY_MAX_LEN: u8 = 32;

fn process_secret<'a>(secret: &str, output: &mut [u8]) -> usize {
    // RFC 6238 Doesn't specify a max length for the secret keys, So I chose 32
    // You can change the max length by modifying the const
    let mut result = [0u8; SECRET_KEY_MAX_LEN];
    let mut inc = 0;
    let offset = match 8 - (secret.len() % 8) {
        8 => 0,
        n => n,
    };

    for c in secret.chars() {
        result[inc] = c.to_ascii_uppercase() as u8;
        inc += 1;
    }
    for _ in 0..offset {
        result[inc] = '=' as u8;
        inc += 1;
    }

    let decoded = match binascii::b32decode(&result[..inc], output).ok() {
        Some(bytes) => bytes,
        None => "".as_bytes(), // TODO: Implement a way to display errors
    };
    decoded.len()
}

fn generate_otp_token(private_key: &str, actual_time: u64) -> u32 {
    let mut output = [0u8; 50];
    let private_key_len = process_secret(private_key, &mut output);

    let interval = actual_time / INTERVAL;
    let msg = interval.to_be_bytes();

    let mut hmac_digest = [0u8; 20];
    hmac_sha1(&output[..private_key_len], &msg, &mut hmac_digest);

    let start = (hmac_digest[19] & 0xF) as usize;
    let bytes: [u8; 4] = hmac_digest[start..start + 4].try_into().unwrap();

    let raw_token = u32::from_be_bytes(bytes);
    ((raw_token & 0x7F_FF_FF_FF) % 1_000_000) as u32
}

fn token_to_bytes(mut token: u32, output: &mut [u8]) {
    let mut inc = 0;
    while token > 0 {
        output[inc] = (token % 10) as u8 + b'0';
        token /= 10;
        inc += 1;
    }
    output.reverse()
}

#[arduino_hal::entry]
fn main() -> ! {
    let dp = arduino_hal::Peripherals::take().unwrap();
    let pins = arduino_hal::pins!(dp);

    let mut led = pins.d13.into_output();
    let mut serial = Usart::new(
        dp.USART0,
        pins.d0,
        pins.d1.into_output(),
        BaudrateArduinoExt::into_baudrate(9600),
    );

    // let token = generate_otp_token("holaaaaamundo", 1680155315);
    let token = generate_otp_token("FS7J22EHLLSOGKUVJKV2XE7FTIX24JAJ", 1680155315);
    let mut output = [0u8; 6];
    token_to_bytes(token, &mut output);

    for b in output {
        serial.write_byte(b);
    }
    serial.write_byte('\n' as u8);

    loop {
        led.toggle();
        delay_ms(1000);
    }
}
